#ifndef __SEZNAM_H__
#define __SEZNAM_H__

#include <memory>
#include "seznam_impl.h"

template<class T, 
	template<class T> class Aloktor = std::allocator,
	class Implementace = seznam_impl<T, Aloktor>
>
class seznam
{
	Implementace *implementace;
public:
	// Zveejovan datov typy
	using typ_hodnot = typename Implementace::typ_hodnot;
	using itertor_seznamu = typename Implementace::itertor_seznamu;
	using konstantn_itertor_seznamu = typename Implementace::konstantn_itertor_seznamu;

	// Konstruktory a destruktor
	seznam() { implementace = new Implementace(); }
	seznam(std::initializer_list<T> seznam_inicializtor);
	seznam(const seznam& s) { implementace = new seznam_impl(*(s.implementace)); }
	seznam(seznam&& s) : seznam() {	proho(s); }
	~seznam() { delete implementace; }

	// Piazovn
	seznam& operator = (const seznam& s) { *(this->implementace) = *(s.implementace); return *this;	}
	seznam& operator = (seznam&& s) { proho(s); return *this; }

	// Stav seznamu
	bool przdn() const noexcept { return implementace->przdn(); }
	std::size_t velikost() const noexcept { return implementace->velikost(); }

	// Pstup k prvkm
	typ_hodnot prvn() { return implementace->prvn(); }
	itertor_seznamu najdi(const T co) { return implementace->najdi_impl<false>(co); }
	konstantn_itertor_seznamu najdi(const T co) const { return implementace->najdi_impl<true>(co); }

	// Vkldn do seznamu
	void vlo_na_konec(const typ_hodnot &co){ implementace->vlo_na_konec(co); }
	void vlo_na_konec(typ_hodnot &&co) { implementace->vlo_na_konec(std::move(co)); }
	void vlo_na_zatek(const typ_hodnot& co) { implementace->vlo_na_zatek(co); }
	void vlo_na_zatek(typ_hodnot&& co) { implementace->vlo_na_zatek(std::move(co)); }
	itertor_seznamu vlo_za(itertor_seznamu kam, T& co) { return implementace->vlo_za(kam, co); }
	itertor_seznamu vlo_za(itertor_seznamu kam, T&& co) { return implementace->vlo_za(kam, std::move(co)); }
	
	// Ostraovn
	void odstra_prvn() { implementace->odstra_prvn(); }
	itertor_seznamu odstra(itertor_seznamu ten) { return implementace->odstra(ten); }
	void vyprzdni() { implementace->vyprzdni(); }
	
	// Pro ladn
	void vypi() { implementace->vypi(); }
	
	// Prohozen obsahu dvou seznam
	void proho(seznam &s) { std::swap(implementace, s.implementace); }
	
	// Itertory
	itertor_seznamu begin() noexcept { return implementace->begin(); };
	konstantn_itertor_seznamu begin() const noexcept { return implementace->cbegin(); };
	konstantn_itertor_seznamu cbegin() const noexcept { return implementace->cbegin(); };
	itertor_seznamu end() noexcept { return implementace->end(); };
	konstantn_itertor_seznamu end() const noexcept { return implementace->cend(); };
	konstantn_itertor_seznamu cend() const noexcept { return implementace->cend(); };
	
	// Spojovn seznam
	seznam operator+(seznam& s) &;
	seznam operator+(seznam&& s) &;
	seznam operator+(seznam& s) &&;
	seznam operator+(seznam&& s) &&;
	seznam operator+=(seznam& s) { *implementace += *s.implementace; return *this; }
	seznam operator+=(seznam&& s) { *implementace += std::move(*s.implementace); return *this; }
private:
	seznam(Implementace* impl): implementace(impl){}
	template<bool lev_r_hodnota, bool prav_r_hodnota>
	seznam opertor_plus_impl(seznam& a);
};

template<class T, template<class T> class Aloktor, class Implementace>
template<bool lev_r_hodnota, bool prav_r_hodnota>
seznam<T, Aloktor, Implementace>
seznam<T, Aloktor, Implementace>::opertor_plus_impl(seznam& a)
{
	seznam<T, Aloktor, Implementace> vsledek;
	Implementace pom;
	if constexpr (!lev_r_hodnota && !prav_r_hodnota)
	{
		pom = *implementace + *a.implementace;
	}
	else if constexpr (!lev_r_hodnota && prav_r_hodnota)
	{
		pom = *implementace + std::move(*a.implementace);
	}
	else if constexpr (lev_r_hodnota && !prav_r_hodnota)
	{
		pom = std::move(*implementace) + *a.implementace;
	}
	else
	{
		pom = std::move(*implementace) + std::move(*a.implementace);
	}
	vsledek.implementace->proho(pom);
	return vsledek;
}

template<class T, template<class T> class Aloktor, class Implementace>
seznam<T, Aloktor, Implementace>::seznam(std::initializer_list<T> seznam_inicializtor)
{
	implementace = new seznam_impl<T, Aloktor>(seznam_inicializtor);
}

template<class T, template<class T> class Aloktor, class Implementace>
seznam<T, Aloktor, Implementace> seznam<T, Aloktor, Implementace>::operator+(seznam & s) &
{
	return opertor_plus_impl<false, false>(s);
}

template<class T, template<class T> class Aloktor, class Implementace>
seznam<T, Aloktor, Implementace> seznam<T, Aloktor, Implementace>::operator+(seznam && s) &
{
	return opertor_plus_impl<false, true>(s);
}

template<class T, template<class T> class Aloktor, class Implementace>
seznam<T, Aloktor, Implementace> seznam<T, Aloktor, Implementace>::operator+(seznam & s) &&
{
	return opertor_plus_impl<true, false>(s);
}

template<class T, template<class T> class Aloktor, class Implementace>
seznam<T, Aloktor, Implementace> seznam<T, Aloktor, Implementace>::operator+(seznam && s) &&
{
	return opertor_plus_impl<true, true>(s);
}
#endif //__SEZNAM_H__
