#ifndef __SEZNAM_IMPL_H__
#define __SEZNAM_IMPL_H__
#include <memory>
#include <stdexcept>
#include <initializer_list>
#include <type_traits>
#include <utility>
#include <iostream>

//using namespace std;

template<class T, template<class T> class Aloktor = std::allocator>
class seznam_impl
{
	struct prvek;
	template<class U = T> struct pomoc;
	template<bool je_konstantn> class itertor_seznamu_impl;
public:
	// Zveejovan datov typy
	using typ_hodnot = T;
	using itertor_seznamu = itertor_seznamu_impl<false>;
	using konstantn_itertor_seznamu = itertor_seznamu_impl<true>;

	// Konstruktory a destruktor
	seznam_impl();
	seznam_impl(std::initializer_list<T> seznam_inicializtor);
	seznam_impl(const seznam_impl& si);
	seznam_impl(seznam_impl&& si);
	~seznam_impl();

	// Piazovn
	seznam_impl& operator=(seznam_impl& si);
	seznam_impl& operator=(seznam_impl&& si);
	// Stav seznamu
	std::size_t velikost() const noexcept { return poet_prvk; }
	bool przdn() const noexcept { return hlava == zarka; }

	// Pstup k prvkm, hledn
	typ_hodnot prvn();
	itertor_seznamu najdi(const T co);
	konstantn_itertor_seznamu najdi(const T co) const;
	template<typename typ_itertoru>
	typ_itertoru najdi_impl(T co) const;

	// Vkldn do seznamu
	void vlo_na_konec(T& co);
	void vlo_na_konec(T&& co);
	void vlo_na_zatek(T& co);
	void vlo_na_zatek(T&& co);
	itertor_seznamu vlo_za(itertor_seznamu &kam, T& co);
	itertor_seznamu vlo_za(itertor_seznamu &kam, T&& co);

	// Odstraovn
	void odstra_prvn();
	void vyprzdni();
	itertor_seznamu odstra(itertor_seznamu ten);
	// Pro ladn
	void vypi();

	// Prohozen obsahu dvou seznam
	void proho(seznam_impl& s);

	// Itertory
	itertor_seznamu begin() noexcept;
	konstantn_itertor_seznamu begin() const noexcept;
	konstantn_itertor_seznamu cbegin() const noexcept;
	itertor_seznamu end() noexcept;
	konstantn_itertor_seznamu end() const noexcept;
	konstantn_itertor_seznamu cend() const noexcept;

	// Spojovn seznam
	seznam_impl operator+(seznam_impl& s) &;
	seznam_impl operator+(seznam_impl && s) &;
	seznam_impl&& operator+(seznam_impl& s) &&;
	seznam_impl&& operator+(seznam_impl&& s) &&;
	seznam_impl& operator +=(seznam_impl& s) { pikopruj(s); return *this; }
	seznam_impl& operator +=(seznam_impl&& s) { pipoj(move(s)); return *this; }

private:
	prvek *hlava, *zarka;
	std::size_t poet_prvk;
	// Odlin implementace nkterch operac pro ukazatele
	pomoc<T> pomcka;

	prvek *pprava_vloen();
	void pikopruj(const seznam_impl& s);
	void pipoj(seznam_impl&& s);
	template<bool r_hodnota>
	itertor_seznamu vlo_za_impl(itertor_seznamu &kam, T&& co);
};

template<class T, template<class T> class Aloktor>
template<class U>
struct seznam_impl<T, Aloktor>::pomoc
{
	void vypi_prvek(prvek *ten)
	{
		std::cout << ten->data << " ";
	}
	void odstra_data(T) {}
};

template<class T, template<class T> class Aloktor>
template<class U>
struct seznam_impl<T, Aloktor>::pomoc<U*>
{
	void vypi_prvek(prvek *ten)
	{
		std::cout << "*" << *(ten->data) << " ";
	}
	void odstra_data(T tato_data) 
	{
		delete tato_data;
	}
};


template<class T, template<class T> class Aloktor>
struct seznam_impl<T, Aloktor>::prvek
{
	T data;
	prvek *dal;
	static Aloktor<prvek> aloktor;

	prvek(T& data, prvek *dal) :data(data), dal(dal) {}
	prvek(T&& data, prvek* dal) : data(std::move(data)), dal(dal) {}

	void *operator new(std::size_t)
	{
		return aloktor.allocate(1);
	}

	void operator delete(void *p)
	{
		aloktor.deallocate((prvek*)p, 1);
	}
};

template<class T, template<class T> class Aloktor>
template<bool je_konstantn>
class seznam_impl<T, Aloktor>::itertor_seznamu_impl
{
	friend class seznam_impl<T, Aloktor>;
	using ukazatel_na_prvek = std::conditional_t<je_konstantn, const prvek*, prvek*>;
	ukazatel_na_prvek ukazatel;
public:
	using iterator_category = std::forward_iterator_tag;
	using value_type = T;
	using difference_type = std::ptrdiff_t;
	using reference = std::conditional_t<je_konstantn, const T&, T&>;
	using pointer = std::conditional_t<je_konstantn, const T*, T*>;
	
	itertor_seznamu_impl(const itertor_seznamu_impl<false>& iter)
		: ukazatel(iter.ukazatel)
	{}

	reference operator *() const
	{
		return ukazatel->data;
	}

	pointer operator->() const
	{
		return &(ukazatel->data);
	}

	itertor_seznamu_impl& operator++()
	{
		if (ukazatel->dal)
		{
			ukazatel = ukazatel->dal;
		}
		return *this;
	}
	itertor_seznamu_impl operator++(int)
	{
		itertor_seznamu_impl pom(*this);
		++*this;
		return pom;
	}
	bool operator ==(const itertor_seznamu_impl& iter) const
	{
		return ukazatel == iter.ukazatel;
	}
	bool operator !=(const itertor_seznamu_impl& iter) const
	{
		return !(*this == iter);
	}
private:
	itertor_seznamu_impl(ukazatel_na_prvek np = nullptr)
		: ukazatel(np)
	{}

	inline ukazatel_na_prvek ukazatel_na_aktuln_prvek() const
	{
		return ukazatel;
	}
};

template<class T, template<class> class Aloktor>
Aloktor<typename seznam_impl<T, Aloktor>::prvek> seznam_impl<T, Aloktor>::prvek::aloktor;


template<class T, template<class> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl()
{
	hlava = zarka = new prvek{ T{}, nullptr };
	poet_prvk = 0;
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(std::initializer_list<T> seznam_inicializtor)
	:seznam_impl()
{
	for (auto x : seznam_inicializtor)
	{
		vlo_na_konec(x);
	}
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(const seznam_impl<T, Aloktor>& si)
	: seznam_impl()
{
	pikopruj(si);
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(seznam_impl<T, Aloktor>&& si)
	: seznam_impl()
{
	proho(si);
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::~seznam_impl()
{
	vyprzdni();
	delete hlava;
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor> & seznam_impl<T, Aloktor>::operator=(seznam_impl<T, Aloktor> & si)
{
	vyprzdni();
	pikopruj(si);
	return *this;
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor> & seznam_impl<T, Aloktor>::operator=(seznam_impl<T, Aloktor> && si)
{
	proho(si);
	return *this;
}

template<class T, template<class T> class Aloktor>
T seznam_impl<T, Aloktor>::prvn()
{
	if(przdn)
	{
		throw std::out_of_range("Odstraovn z przdnho seznamu");
	}
	return hlava->data;
}

template<class T, template<class T> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::najdi(const T co)
{
	//return itertor_seznamu(najdi_ukazatel(co));
	return najdi_impl<itertor_seznamu>(co);
}

template<class T, template<class T> class Aloktor>
typename seznam_impl<T, Aloktor>::konstantn_itertor_seznamu seznam_impl<T, Aloktor>::najdi(const T co) const
{
	//return konstantn_itertor_seznamu(najdi_ukazatel(co));
	return najdi_impl<konstantn_itertor_seznamu>(co);
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_konec(T& co)
{
	prvek *pom = pprava_vloen();
	new(&pom->data) T(co);
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_konec(T&& co)
{
	prvek *pom = pprava_vloen();
	new(&pom->data) T(std::move(co));
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::prvek *seznam_impl<T, Aloktor>::pprava_vloen()
{
	prvek* pom = zarka;
	poet_prvk++;
	(pom->data).~T();
	zarka = zarka->dal = new prvek{ T{}, nullptr };
	return pom;
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::odstra_prvn()
{
	if (przdn())
	{
		throw std::out_of_range("Odstraovn z przdnho seznamu");
	}
	poet_prvk--;
	auto pom = hlava;
	hlava = hlava->dal;
	pomcka.odstra_data(pom->data);
	delete pom;
}

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::vyprzdni()
{
	while (!przdn())
		odstra_prvn();
}

template<class T, template<class T> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::odstra(typename seznam_impl<T, Aloktor>::itertor_seznamu kter)
{
	prvek* ten = kter.ukazatel_na_aktuln_prvek();
	if (ten == zarka)
	{
		throw std::out_of_range("Odstraovna zarka seznamu");
	}
	poet_prvk--;
	prvek* nsledujc = ten->dal;
	if (nsledujc == zarka)
	{
		delete zarka;
		zarka = ten;
		zarka->dal = nullptr;
	}
	else
	{
		*ten = *nsledujc;
		delete nsledujc;
	}
	return itertor_seznamu{ nsledujc };
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_zatek(T& co)
{
	hlava = new prvek{ co, hlava };
	poet_prvk++;
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_zatek(T&& co)
{
	hlava = new prvek{ std::move(co), hlava };
	poet_prvk++;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::vlo_za(typename seznam_impl<T, Aloktor>::itertor_seznamu & kam, T& co)
{
	return vlo_za_impl<false>(kam, co);
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::vlo_za(typename seznam_impl<T, Aloktor>::itertor_seznamu & kam, T&& co)
{
	return vlo_za_impl<true>(kam, std::move(co));
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::vypi()
{
	std::cout << "{ ";
	for (prvek *pom = hlava; pom != zarka; pom = pom->dal)
	{
		pomcka.vypi_prvek(pom);
	}
	std::cout << "}";
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::proho(seznam_impl<T, Aloktor>& s)
{
	std::swap(hlava, s.hlava);
	std::swap(zarka, s.zarka);
	std::swap(poet_prvk, s.poet_prvk);
}

template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::pikopruj(const seznam_impl<T, Aloktor>& s)
{
	for (auto pom = s.hlava; pom != s.zarka; pom = pom->dal)
	{
		vlo_na_konec(pom->data);
	}
}

//template<class T, template<class> class Aloktor>
//typename seznam_impl<T, Aloktor>::prvek * seznam_impl<T, Aloktor>::najdi_ukazatel(const T co) const
//{
//	for (prvek *pom = hlava; pom != zarka; pom = pom->dal)
//	{
//		if (pom->data == co)
//		{
//			return pom;
//		}
//	}
//	return zarka;
//}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::begin() noexcept
{
	return hlava;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::konstantn_itertor_seznamu seznam_impl<T, Aloktor>::begin() const noexcept
{
	return hlava;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::konstantn_itertor_seznamu seznam_impl<T, Aloktor>::cbegin() const noexcept
{
	return hlava;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::end() noexcept
{
	return zarka;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::konstantn_itertor_seznamu seznam_impl<T, Aloktor>::end() const noexcept
{
	return zarka;
}

template<class T, template<class> class Aloktor>
typename seznam_impl<T, Aloktor>::konstantn_itertor_seznamu seznam_impl<T, Aloktor>::cend() const noexcept
{
	return zarka;
}

template<class T, template<class> class Aloktor>
seznam_impl<T, Aloktor> seznam_impl<T, Aloktor>::operator+(seznam_impl & s) &
{
	seznam_impl<T, Aloktor> vsledek(*this);
	vsledek.pikopruj(s);
	return vsledek;
}

template<class T, template<class> class Aloktor>
seznam_impl<T, Aloktor>&& seznam_impl<T, Aloktor>::operator+(seznam_impl & s) &&
{
	pikopruj(s);
	return move(*this);
}

template<class T, template<class> class Aloktor>
seznam_impl<T, Aloktor> seznam_impl<T, Aloktor>::operator+(seznam_impl && s) &
{
	seznam_impl<T, Aloktor> vsledek(*this);
	vsledek.pipoj(move(s));
	return vsledek;
}

template<class T, template<class> class Aloktor>
seznam_impl<T, Aloktor>&& seznam_impl<T, Aloktor>::operator+(seznam_impl && s) &&
{
	pipoj(move(s));
	return move(*this);
}



template<class T, template<class> class Aloktor>
void seznam_impl<T, Aloktor>::pipoj(seznam_impl<T, Aloktor>&& s)
{
	if (s.przdn()) return;
	zarka->data = s.hlava->data;
	zarka->dal = s.hlava->dal;
	zarka = s.zarka;
	delete s.hlava;
	s.hlava = s.zarka = new prvek{ T{}, nullptr };
	poet_prvk += s.poet_prvk;
	s.poet_prvk = 0;
}

template<class T, template<class> class Aloktor>
template<typename typ_itertoru>
inline typ_itertoru seznam_impl<T, Aloktor>::najdi_impl(T co) const
{
	for (auto pom = hlava; pom != zarka; pom = pom->dal)
	{
		if (pom->data == co)
		{
			return typ_itertoru{ pom };
		}
	}
	return typ_itertoru{ zarka };
}

template<class T, template<class> class Aloktor>
template<bool r_hodnota>
typename seznam_impl<T, Aloktor>::itertor_seznamu seznam_impl<T, Aloktor>::vlo_za_impl(typename seznam_impl<T, Aloktor>::itertor_seznamu &kam, T&& co)
{
	prvek* pom = kam.ukazatel_na_aktuln_prvek();
	if (pom == zarka)
	{
		throw std::out_of_range{ "Vkldn za zarku seznamu" };
	}
	poet_prvk++;
	prvek* nsledujc = pom->dal;
	
	if constexpr(r_hodnota)
	{
		pom->dal = new prvek(std::move(co), nsledujc);
	}
	else
	{
		pom->dal = new prvek(co, nsledujc);
	}
	return itertor_seznamu{ pom->dal };
}

#endif //__SEZNAM_IMPL_H__

